from __future__ import annotations

from bigsmiles.config import Config


class Atom:
    __slots__ = ["id_", "element", "isotope", "stereo", "hydrogens", "charge", "valence",
                 "organic", "bonds", "possible_valence", "_default_valance", "__dict__", "parent"]
    _tree_print_repr = True

    def __init__(self,
                 id_: int,
                 element: str,
                 isotope: int | None = None,
                 stereo: str = '',
                 hydrogens: int = 0,
                 charge: int = 0,
                 valence: int = None,
                 parent: BigSMILES | Branch | StochasticFragment | None = None,
                 **kwargs
                 ):
        self.id_ = id_
        self.element = element
        self.isotope = isotope
        self.stereo = stereo
        self.hydrogens = hydrogens
        self.charge = charge
        self.possible_valence: tuple[int] = Config.get_atom_possible_valence(element)
        if valence is None:
            self._default_valance: bool = True
            self.valence = self.possible_valence[0]
        else:
            self._default_valance: bool = False
            self.valence = valence
        self.organic = True if element in Config.organics else False

        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

        self.bonds = []

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return str(self) + "{" + str(self.id_) + "}"

    @property
    def bond_capacity(self) -> int:
        return self.valence - self.hydrogens + self.charge

    @property
    def number_of_bonds(self) -> int:
        return sum((bond.bond_order for bond in self.bonds)) + self.hydrogens

    @property
    def bonds_available(self) -> int:
        return self.bond_capacity - self.number_of_bonds

    @property
    def ring_indexes(self) -> list[int]:
        ring_index = []
        for bond in self.bonds:
            if bond.ring_id is not None:
                ring_index.append(bond.ring_id)

        return ring_index

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False) -> str:
        text = self.element
        bracket_flag = False

        if self.element == "H":
            bracket_flag = True

        if self.stereo:
            text += self.stereo
            bracket_flag = True

        if show_hydrogens or self.hydrogens > 0:
            if self.hydrogens > 0 or self.bonds_available != 0:
                text += f"H{self.bonds_available + self.hydrogens}"
                bracket_flag = True

        if self.charge > 0:
            text += f"+{self.charge if self.charge > 1 else ''}"
            bracket_flag = True
        elif self.charge < 0:
            text += f"{self.charge}"
            bracket_flag = True

        if self.isotope is not None:
            text = str(self.isotope) + text
            bracket_flag = True

        if bracket_flag:
            text = "[" + text + "]"

        if self.ring_indexes:
            for bond in self.bonds:
                if bond.ring_id is not None:
                    if bond.ring_id > 9:
                        ring_id = "%" + str(bond.ring_id)
                    else:
                        ring_id = str(bond.ring_id)
                    text += bond.symbol + ring_id

        return text

    def _increase_valance(self, requested_valance_increase: int = 0) -> bool:
        " If default valance, try to increase when an additional bond added that needs it. And does the increase!! "
        if not self._default_valance:
            return False

        for valence in self.possible_valence:
            if valence > self.valence:
                old_valance = self.valence
                self.valence = valence
                if self.bonds_available < requested_valance_increase:
                    # increase was not enough for requested valance increase; set valence back to original value
                    self.valence = old_valance
                    continue

                return True

        return False

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


bond_mapping = {
    "": 1,
    "=": 2,
    "#": 3
}


class Bond:
    __slots__ = ["id_", "symbol", "atom1", "atom2", "ring_id", "__dict__", "parent"]
    _tree_print_repr = True

    def __init__(self,
                 symbol: str,
                 atom1: Atom | BondDescriptorAtom | StochasticObject,
                 atom2: Atom | BondDescriptorAtom | StochasticObject | None = None,
                 id_: int = None,
                 ring_id: int = None,
                 parent: BigSMILES | Branch | StochasticFragment | None = None,
                 **kwargs
                 ):
        self.id_ = id_
        self.symbol = symbol
        self.atom1 = atom1
        self.atom2 = atom2
        self.ring_id = ring_id
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False):
        return Config.add_color(self.symbol, 'Blue', skip_color)

    def __iter__(self):
        return iter((self.atom1, self.atom2))

    @property
    def bond_order(self) -> int:
        return bond_mapping[self.symbol]

    @bond_order.setter
    def bond_order(self, count: int):
        for k, v in bond_mapping.items():
            if v == count:
                self.symbol = k
                return

        raise ValueError(f"Invalid bond_order. \nGiven: {count} \n Acceptable: {bond_mapping}")

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


class BondDescriptor:
    __slots__ = ["parent", "descriptor", "index_", "instances", "__dict__"]

    def __init__(self, parent: StochasticObject, descriptor: str, index_: int, **kwargs):
        self.parent = parent
        self.descriptor = descriptor
        self.index_ = index_
        self.instances = []

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False):
        if self.index_ == 1 and not Config.show_bond_descriptor_zero_index:
            index = ""
        else:
            index = self.index_
        return "[" + self.descriptor + str(index) + "]"

    @property
    def symbol(self) -> str:
        return self.descriptor + str(self.index_)

    def is_pair(self, bd: BondDescriptor) -> bool:
        """ Returns true if symbols are <> and index match. """
        if self.index_ == bd.index_:
            if (self.descriptor == "<" and bd.descriptor == ">") or \
                    (self.descriptor == ">" and bd.descriptor == "<"):
                return True

        return False

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


class BondDescriptorAtom:
    _tree_print_repr = True
    __slots__ = ["descriptor", "id_", "bond", "__dict__", "parent"]

    def __init__(self,
                 bond_descriptor: BondDescriptor,
                 id_: int = None,
                 parent: BigSMILES | Branch | StochasticFragment | None = None,
                 **kwargs):
        self.descriptor = bond_descriptor
        bond_descriptor.instances.append(self)
        self.id_ = id_
        self.bond = None
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False):
        return Config.add_color(self.descriptor.to_string(show_hydrogens, print_repr), 'Green', skip_color)

    @property
    def root(self) -> BigSMILES:
        return self.parent.root


class Branch:
    _tree_print_repr = False
    __slots__ = ["nodes", "id_", "parent", "__dict__"]

    def __init__(self, parent: BigSMILES | StochasticFragment | Branch, id_: int = None, **kwargs):
        self.nodes: list[Atom | Bond | Branch | StochasticObject | BondDescriptorAtom] = []
        self.id_ = id_
        self.parent = parent

        if kwargs:
            for k, v in kwargs.items():
                setattr(self, k, v)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False):
        text = "".join(
            node.to_string(show_hydrogens, print_repr) for node in self.nodes
        )

        return "(" + text + ")"

    @property
    def in_stochastic_object(self) -> bool:
        return self.parent.in_stochastic_object

    @property
    def root(self) -> BigSMILES:
        return self.parent.root

    def _get_id(self) -> int:
        return self.root._get_id()


class StochasticFragment:
    _tree_print_repr = False
    __slots__ = ["nodes", "id_", "parent", "bonding_descriptors", "rings", "__dict__"]

    def __init__(self, parent: StochasticObject, id_: int = None):
        self.nodes: list[Atom | Bond | BondDescriptorAtom | Branch | StochasticObject] = []
        self.rings: list[Bond] = []
        self.bonding_descriptors: list[BondDescriptor] = []
        self.id_ = id_
        self.parent = parent

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False):
        return "".join(node.to_string(show_hydrogens, print_repr, skip_color) for node in self.nodes)

    @property
    def in_stochastic_object(self) -> bool:
        return self.parent.in_stochastic_object

    @property
    def root(self) -> BigSMILES:
        return self.parent.root

    def _get_id(self) -> int:
        return self.root._get_id()


class StochasticObject:
    _tree_print_repr = False
    __slots__ = ["nodes", "bonding_descriptors", "id_", "parent", "end_group_left", "end_group_right",
                 "bond_left", "bond_right", "__dict__"]

    def __init__(self, parent: BigSMILES | StochasticFragment | Branch, id_: int = None):
        self.nodes: list[StochasticFragment] = []
        self.bonding_descriptors: list[BondDescriptor] = []
        self.end_group_left: BondDescriptorAtom | None = None
        self.end_group_right: BondDescriptorAtom | None = None
        self.id_ = id_
        self.bond_left: Bond | None = None
        self.bond_right: Bond | None = None
        self.parent = parent

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False):
        text = Config.add_color("{", "Red", skip_color)
        text += self.end_group_left.to_string(show_hydrogens, print_repr, skip_color)
        text += ",".join(node.to_string(show_hydrogens, print_repr, skip_color) for node in self.nodes)
        text += self.end_group_right.to_string(show_hydrogens, print_repr, skip_color)
        return text + Config.add_color("}", "Red", skip_color)

    @property
    def implicit_endgroups(self) -> bool:
        """ Returns true if one or more are implicit """
        return True if self.end_group_left.descriptor.descriptor == "" or \
                       self.end_group_right.descriptor.descriptor == "" else False

    @property
    def in_stochastic_object(self) -> bool:
        return True

    @property
    def root(self) -> BigSMILES:
        return self.parent.root

    def _get_id(self) -> int:
        return self.root._get_id()


def contains_stochastic_object(nodes: Atom | Bond | StochasticObject | Branch):
    for node in nodes:
        if isinstance(node, StochasticObject):
            return True
        if isinstance(node, Branch):
            result = contains_stochastic_object(node.nodes)
            if result:
                return True

    return False


class BigSMILES:
    _tree_print_repr = False
    __slots__ = ["nodes", "atoms", "bonds", "rings", "__dict__"]

    def __init__(self, input_text: str = None):
        self.nodes: list[Atom | Bond | StochasticObject | Branch] = []
        self.atoms: list[Atom | BondDescriptorAtom] = []  # includes atoms in sub-objects
        self.bonds: list[Bond] = []  # includes bonds in sub-objects
        self.rings: list[Bond] = []  # does not include rings in sub-objects

        self.__id = 0
        # parse input string
        if input_text:
            from bigsmiles.parse_bigsmiles_str import parse_bigsmiles_str
            parse_bigsmiles_str(input_text, self)

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string(print_repr=True, skip_color=True)

    def __getitem__(self, obj: int | slice) -> Atom | list[Atom]:
        return self.atoms[obj]

    def __iter__(self) -> list[Atom]:
        return self.atoms

    def __bool__(self):
        if self.nodes:
            return True
        else:
            return False

    def to_string(self, show_hydrogens: bool = False, print_repr: bool = False, skip_color: bool = False) -> str:
        return "".join(node.to_string(show_hydrogens, print_repr, skip_color) for node in self.nodes)

    @property
    def in_stochastic_object(self) -> bool:
        return False

    @property
    def contains_stochastic_object(self) -> bool:
        return contains_stochastic_object(self.nodes)

    @property
    def root(self) -> BigSMILES:
        return self

    def _get_id(self) -> int:
        id_ = self.__id
        self.__id += 1
        return id_

    def print_tree(self, show_object_label: bool = True, print_repr: bool = False):
        """
        prints a tree representation of the parsed bigsmiles

        Parameters
        ----------
        show_object_label: bool
            show object labels
        print_repr: bool
            use repr() instead of str()

        """
        from bigsmiles.tree_to_string import tree_to_string  # here to avoid circular imports
        print(tree_to_string(self, show_object_label, print_repr))



import re
from functools import wraps

from bigsmiles.config import Config
from bigsmiles.errors import BigSMILESError
from bigsmiles.constructor import *

ATOM_PATTERN = re.compile(
    r"(?P<isotope>\d{1,3})?" +
    r'(?P<element>' + "|".join(Config.elements_ordered) + "|".join(Config.aromatic) + ')' +
    r"(?P<stereo>@{0,2})?" +
    r'(?P<hydrogens>H\d?)?' +
    r'(?P<charge>[-|+]{1,3}\d?)?'
)


def atom_symbol_to_attributes(symbol: str) -> dict:
    return ATOM_PATTERN.match(symbol).groupdict()


DEFAULT_BONDING_DESCRIPTROR_INDEX = 1


def process_bonding_descriptor_symbol(symbol: str) -> tuple[str, int]:
    symbol = symbol.replace("[", "").replace("]", "")
    if not symbol:
        return symbol, DEFAULT_BONDING_DESCRIPTROR_INDEX

    if symbol[-1].isdigit():
        return symbol[0], int(symbol[-1])

    return symbol, DEFAULT_BONDING_DESCRIPTROR_INDEX


def in_stochastic_object(func):
    """ Decorator to ensure function call only occurs within stochastic object."""

    @wraps(func)
    def _in_stochastic_object(*args, **kwargs):
        if not args[0].in_stochastic_object:
            raise BigSMILESError("Must in stochastic object.")
        return func(*args, **kwargs)

    return _in_stochastic_object


def add_atom_str(parent: ParentType, symbol: str) -> ParentType:
    return add_atom(parent, **atom_symbol_to_attributes(symbol))


@in_stochastic_object
def add_bonding_descriptor_str(parent: ParentType, bd_symbol: str) -> ParentType:
    return add_bonding_descriptor(parent, *process_bonding_descriptor_symbol(bd_symbol))


def add_bond_atom_pair_str(parent: ParentType, bond_symbol: str,
                           atom_symbol: str) -> ParentType:
    return add_bond_atom_pair(parent, bond_symbol, **atom_symbol_to_attributes(atom_symbol))


@in_stochastic_object
def add_bond_bonding_descriptor_pair_str(parent: ParentType, bond_symbol: str,
                                         bd_symbol: str) -> ParentType:
    return add_bond_bonding_descriptor_pair(parent, bond_symbol, *process_bonding_descriptor_symbol(bd_symbol))


def open_stochastic_object_str(parent: ParentType, bd_symbol: str) -> ParentType:
    return open_stochastic_object(parent, *process_bonding_descriptor_symbol(bd_symbol))


def open_stochastic_object_with_bond_str(parent: ParentType, bond_symbol: str,
                                         bd_symbol: str) -> ParentType:
    return open_stochastic_object_with_bond(parent, bond_symbol, *process_bonding_descriptor_symbol(bd_symbol))


def close_stochastic_object_str(parent: ParentType, bd_symbol: str) -> ParentType:
    return close_stochastic_object(parent, *process_bonding_descriptor_symbol(bd_symbol))


@in_stochastic_object
def open_stochastic_fragment(parent) -> ParentType:
    return open_stochastic_fragment(parent)


@in_stochastic_object
def close_open_stochastic_fragment_str(parent) -> ParentType:
    return close_open_stochastic_fragment(parent)


@in_stochastic_object
def close_stochastic_fragment_str(parent) -> ParentType:
    return close_stochastic_fragment(parent)



from bigsmiles.errors import BigSMILESError
from bigsmiles.bigsmiles import Atom, Bond, BondDescriptor, Branch, StochasticFragment, StochasticObject, \
    BigSMILES, BondDescriptorAtom
from bigsmiles.validation import post_validation
from bigsmiles.syntax_fixes import run_syntax_fixes

ParentType = BigSMILES | Branch | StochasticFragment | StochasticObject


## Utility functions for construction ##
######################################################################################################################
def add_bond_to_connected_objects(bond: Bond):
    """ Adds bonds to Atoms, and BondDescriptorAtom. """
    for obj in bond:
        if isinstance(obj, Atom):
            if obj.bonds_available < bond.bond_order:
                if obj._increase_valance(bond.bond_order):
                    continue
                raise BigSMILESError("Too many bonds trying to be made.", str(obj))
            obj.bonds.append(bond)
        elif isinstance(obj, BondDescriptorAtom):
            obj.bond = bond
        elif isinstance(obj, StochasticObject):  # left bond add with
            obj.bond_right = bond


def add_bond_descriptor_to_stochastic_fragment(stoch_frag: StochasticFragment, loop_obj: Branch = None):
    if loop_obj is None:
        loop_obj = stoch_frag

    for obj in loop_obj.nodes:
        if isinstance(obj, BondDescriptorAtom) and obj.descriptor not in stoch_frag.bonding_descriptors:
            stoch_frag.bonding_descriptors.append(obj.descriptor)
        if isinstance(obj, Branch):
            add_bond_descriptor_to_stochastic_fragment(stoch_frag, obj)  # recursive


def get_common_bond(atom1: Atom, atom2: Atom) -> Bond | None:
    # Check if bond already between two atoms
    for bond in atom1.bonds:
        if bond in atom2.bonds:
            return bond

    return None  # No common bond found


def remove_partial_ring(parent: BigSMILES | Branch | StochasticFragment, ring_id: int):
    for i, ring in enumerate(parent.rings):
        if ring.ring_id == ring_id:
            break
    else:
        return

    bond = parent.rings.pop(i)
    parent.bonds.remove(bond)
    if hasattr(parent, "parent"):
        parent.parent.bonds.remove(bond)


def insert_obj_into_list(list_, obj, objet_to_insert):
    for i, item in enumerate(list_):
        if item == obj:
            flag = True
            while flag:
                if isinstance(list_[i + 1], Branch):
                    i += 1
                    continue

                flag = False

            list_.insert(i + 1, objet_to_insert)
            break


def get_prior(obj: BigSMILES | StochasticObject | StochasticFragment | Branch, types_: tuple,
              flag: bool = False) -> Atom:
    if obj.nodes:
        for node in reversed(obj.nodes):  # no atom can ever have more than 10 bonds
            if type(node) in types_:
                return node

        raise BigSMILESError("Trying to make a bond to a prior atom that isn't there.??")

    elif not flag:
        return get_prior(obj.parent, types_, flag=True)

    raise BigSMILESError("Bond attempted to be made to that has nothing to bond back to.")


def get_ring_parent(parent) -> BigSMILES | StochasticFragment:
    """ Recursive """
    if hasattr(parent, "rings"):
        return parent

    return get_ring_parent(parent.parent)


def run_validation(parent):
    """ Run various validations. """
    post_validation(parent)


def exit_(parent):
    if not isinstance(parent, BigSMILES):
        raise BigSMILESError(f"{type(parent)} is missing closing symbol.")
    run_syntax_fixes(parent)


## Functions for constructing a bigsmiles sequentially or step by step ##
#######################################################################################################################
def add_atom(
        parent: ParentType,
        element: str,
        isotope: int | None = None,
        stereo: str = '',
        hcount: int = 0,
        charge: int = 0,
        valance: int = None,
        **kwargs
) -> ParentType:
    atom = Atom(parent._get_id(), element, isotope, stereo, hcount, charge, valance,
                parent=parent, **kwargs)
    parent.nodes.append(atom)
    parent.root.atoms.append(atom)

    return parent


def add_bond(
        parent: ParentType,
        bond_symbol: str,
        atom1: Atom | BondDescriptorAtom,
        atom2: Atom | BondDescriptorAtom | StochasticObject | None,
        **kwargs
) -> ParentType:
    bond = Bond(bond_symbol, atom1, atom2, parent._get_id(), parent=parent, **kwargs)
    parent.nodes.append(bond)
    parent.root.bonds.append(bond)
    add_bond_to_connected_objects(bond)

    return parent


def add_ring(parent: ParentType, ring_id: int, bond_symbol: str = "", **kwargs) -> ParentType:
    # check if ring_id already exists
    ring_parent = get_ring_parent(parent)
    for ring in ring_parent.rings:
        if ring.ring_id == ring_id:
            if ring.atom2 is not None:
                raise BigSMILESError(f"Ring already formed for ring id {ring_id}.")
            atom2 = get_prior(parent, (Atom,))

            bond = get_common_bond(ring.atom1, atom2)
            if bond is None:
                # Add atom2
                ring.atom2 = atom2
                add_bond_to_connected_objects(ring)
            else:
                # Check if ring already between two atoms and just increase bond order
                bond.bond_order += 1
                remove_partial_ring(ring_parent, ring_id)

            return parent

    # Make new ring_id
    bond = Bond(bond_symbol, get_prior(parent, (Atom,)), None, parent._get_id(), ring_id,
                parent=parent, **kwargs)
    parent.root.bonds.append(bond)
    ring_parent.rings.append(bond)

    return parent


def add_ring_from_atoms(parent: ParentType, atom1: Atom, atom2: Atom, bond_symbol: str = "", **kwargs) \
        -> ParentType:
    ring_parent = get_ring_parent(parent)
    bond = get_common_bond(atom1, atom2)  # Check if ring already between two atoms
    if bond is None:
        # Make new ring_id
        bond = Bond(bond_symbol, atom1, atom2, parent._get_id(), len(parent.root.rings) + 1,
                    parent=parent, **kwargs)
        parent.root.bonds.append(bond)
        ring_parent.rings.append(bond)
        add_bond_to_connected_objects(bond)
    else:
        # increase bond order of existing bond
        bond.bond_order += 1

    return parent


def _get_bonding_descriptor_atom(parent, descriptor: str, index_: int, **kwargs) -> BondDescriptorAtom:
    bd = _get_bonding_descriptor(parent, descriptor, index_)
    return BondDescriptorAtom(bd, parent._get_id(), parent=parent, **kwargs)


def _get_bonding_descriptor(parent, descriptor: str, index_: int) -> BondDescriptor:
    stoch_obj = _get_current_stochastic_object(parent)

    # check if bd in there already
    for bd in stoch_obj.bonding_descriptors:
        if descriptor == bd.descriptor and index_ == bd.index_:
            return bd

    # create new bonding descriptor
    new_bd = BondDescriptor(stoch_obj, descriptor, index_)
    stoch_obj.bonding_descriptors.append(new_bd)
    return new_bd


def _get_current_stochastic_object(parent) -> StochasticObject:
    """ Recursive"""
    if isinstance(parent, StochasticObject):
        return parent

    if hasattr(parent, "parent"):
        return _get_current_stochastic_object(parent.parent)
    else:
        raise BigSMILESError("No stochastic object found.")


def add_bonding_descriptor(parent, descriptor: str, index_: int, **kwargs) -> BondDescriptorAtom:
    """ [<], [>], [$], [$1], [>2], ... """
    bd_atom = _get_bonding_descriptor_atom(parent, descriptor, index_, **kwargs)
    parent.nodes.append(bd_atom)
    parent.root.atoms.append(bd_atom)
    return parent


def add_bond_atom_pair(
        parent,
        bond_symbol: str,
        element: str,
        isotope: int | None = None,
        stereo: str = '',
        hcount: int = 0,
        charge: int = 0,
        valance: int = None,
        atom_kwargs: dict = None,
        bond_kwargs: dict = None
) -> tuple[Bond, Atom]:
    atom_kwargs = atom_kwargs if atom_kwargs else {}
    bond_kwargs = bond_kwargs if bond_kwargs else {}

    atom = Atom(parent._get_id(), element, isotope, stereo, hcount, charge, valance,
                parent=parent, **atom_kwargs)
    prior_atom = get_prior(parent, (Atom, BondDescriptorAtom, StochasticObject))
    add_bond(parent, bond_symbol, prior_atom, atom, **bond_kwargs)
    parent.nodes.append(atom)
    parent.root.atoms.append(atom)

    return parent


def add_bond_bonding_descriptor_pair(
        parent,
        bond_symbol: str,
        descriptor: str,
        index_: int,
        bond_descriptor_kwargs: dict = None,
        bond_kwargs: dict = None
):
    bond_descriptor_kwargs = bond_descriptor_kwargs if bond_descriptor_kwargs else {}
    bond_kwargs = bond_kwargs if bond_kwargs else {}

    bd_atom = _get_bonding_descriptor_atom(parent, descriptor, index_, **bond_descriptor_kwargs)
    prior_atom = get_prior(parent, (Atom, BondDescriptorAtom, StochasticObject))
    add_bond(parent, bond_symbol, prior_atom, bd_atom, **bond_kwargs)
    parent.nodes.append(bd_atom)
    parent.root.atoms.append(bd_atom)
    return parent


def open_branch(parent, **kwargs):
    if len(parent.nodes) == 0 and hasattr(parent, 'parent') and isinstance(parent.parent, Branch):
        raise BigSMILESError("BigSMILES string or branch can't start with branch")

    branch = Branch(parent, parent._get_id(), **kwargs)
    parent.nodes.append(branch)

    return branch


def close_branch(parent):
    if not isinstance(parent, Branch):
        raise BigSMILESError("Error closing branch. Possible issues: "
                             "\n\tClosing a branch with another intermediate node started."
                             "\n\tNo starting branch symbol.")

    # check if branch is empty; if so remove it
    if len(parent.nodes) == 0:
        parent.parent.nodes.pop()

    return parent.parent


def open_stochastic_object(parent, descriptor: str, index_: int, **kwargs) -> StochasticFragment:
    stoch_obj = StochasticObject(parent, parent._get_id())
    new_bd = BondDescriptor(stoch_obj, descriptor, index_)
    stoch_obj.bonding_descriptors.append(new_bd)
    stoch_obj.end_group_left = BondDescriptorAtom(new_bd, parent._get_id(),
                                                  parent=parent, **kwargs)
    parent.nodes.append(stoch_obj)

    fragment = StochasticFragment(stoch_obj, stoch_obj._get_id())
    stoch_obj.nodes.append(fragment)

    return fragment


def open_stochastic_object_with_bond(parent, bond_symbol: str, descriptor: str, index_: int, **kwargs) \
        -> StochasticFragment:
    stoch_obj = StochasticObject(parent, parent._get_id())
    new_bd = BondDescriptor(stoch_obj, descriptor, index_)
    stoch_obj.bonding_descriptors.append(new_bd)
    stoch_obj.end_group_left = BondDescriptorAtom(new_bd, parent._get_id(),
                                                  parent=parent, **kwargs)

    prior_atom = get_prior(parent, (Atom, BondDescriptor, StochasticObject))
    bond = add_bond(parent, bond_symbol, prior_atom, stoch_obj)
    stoch_obj.bond_left = bond

    parent.nodes.append(stoch_obj)

    return open_stochastic_fragment(stoch_obj)


def close_stochastic_object(parent, descriptor: str, index_: int):
    if not isinstance(parent, StochasticObject):
        raise BigSMILESError("Error closing StochasticObject. Possible issues: "
                             "\n\tClosing a StochasticObject with another intermediate node started."
                             "\n\tNo starting StochasticObject symbol.")

    parent.end_group_right = _get_bonding_descriptor_atom(parent, descriptor, index_)

    # TODO: add [H] and bond if no atom before or after
    return parent.parent


def open_stochastic_fragment(parent) -> StochasticFragment:
    stoch_frag = StochasticFragment(parent, parent._get_id())
    parent.nodes.append(stoch_frag)
    return stoch_frag


def close_open_stochastic_fragment(parent) -> StochasticFragment:
    if not isinstance(parent, StochasticFragment):
        raise BigSMILESError("Stochastic seperator can only follow a stochastic fragments.")

    parent = close_stochastic_fragment(parent)
    return open_stochastic_fragment(parent)


def close_stochastic_fragment(parent):
    if not isinstance(parent, StochasticFragment):
        raise BigSMILESError("Error StochasticFragment branch. Possible issues: "
                             "\n\tClosing a branch with another intermediate node started.")

    add_bond_descriptor_to_stochastic_fragment(parent)

    # check for at-least one bonding descriptor
    if not parent.bonding_descriptors:
        raise BigSMILESError(f"No bonding descriptor in the stochastic fragment. ({repr(parent)})")

    return parent.parent


## Methods that don't use state and are for editing ##
###################################################################################################################
def append_bigsmiles_fragment(parent, bond_symbol: str | None, bigsmiles_: BigSMILES):
    if bond_symbol is not None:
        if not isinstance(bigsmiles_.nodes[0], Atom | StochasticObject):
            raise BigSMILESError("First node must be an 'Atom' to added fragment.")

        atom = bigsmiles_.nodes[0]
        prior_atom = get_prior(parent, (Atom, BondDescriptorAtom, StochasticObject))
        add_bond(parent, bond_symbol, prior_atom, atom)
        # stack[-1].nodes.append(atom)
        # bigsmiles.atoms.append(atom)

    # append bigsmiles
    # re-direct 'parent' to new bigsmiles object
    for node in bigsmiles_.nodes:
        if hasattr(node, 'parent'):
            node.parent = parent
    # re-numbering
    # re_number_rings(parent, set())
    # re_number_nodes(bigsmiles_)

    # append fragment
    parent.nodes += bigsmiles_.nodes
    parent.root.atoms += bigsmiles_.atoms
    parent.root.bonds += bigsmiles_.bonds
    parent.root.rings += bigsmiles_.rings


def attach_bigsmiles_branch(parent, bond_symbol: str | None, bigsmiles_: BigSMILES, index_: int):
    if index_ > len(parent.root.atoms) - 1:
        raise BigSMILESError(f"Branch can't be added. 'index' outside atom list."
                             f"\n current smiles: {parent.root} \n desired atom index: {index_}")

    # re-direct 'parent' to new bigsmiles object
    for node in bigsmiles_.nodes:
        if hasattr(node, 'parent'):
            node.parent = parent
    # # re-numbering
    # re_number_rings(parent, set())
    # re_number_nodes(bigsmiles_)

    atom = parent.root.atoms[index_]

    # make branch
    branch = Branch(parent, parent._get_id())
    branch.nodes = bigsmiles_.nodes
    parent.root.atoms += bigsmiles_.atoms
    parent.root.bonds += bigsmiles_.bonds

    # make bond
    bond = Bond(bond_symbol, atom, branch.nodes[0], parent._get_id(), parent=parent)
    branch.nodes.insert(0, bond)
    parent.root.bonds.append(bond)
    add_bond_to_connected_objects(bond)

    # insert branch by put it after all the other branches
    insert_obj_into_list(parent.root.nodes, atom, branch)


# def re_number_rings(obj, seen: set[Bond]):
#     """ Recursive renumbering of rings. """
#     for node in obj.nodes:
#         if isinstance(node, Bond) and node.ring_id is not None and node not in seen:
#             obj.root.rings.append(node)
#             node.ring_id = len(obj.root.rings)
#             seen.add(node)
#         if hasattr(node, 'nodes'):
#             re_number_rings(node, seen)
#
#
# def re_number_nodes(obj):
#     """ Recursive renumbering 'id_'. """
#     for node in obj.nodes:
#         if isinstance(node, Atom):
#             node.id_ = parent._get_id()
#         elif isinstance(node, Bond):
#             node.id_ = parent._get_id()
#         elif isinstance(node, BondDescriptorAtom):
#             node.id_ = parent._get_id()
#         elif isinstance(node, Branch):
#             node.id_ = parent._get_id()
#             re_number_nodes(node)
#         elif isinstance(node, StochasticFragment):
#             node.id_ = parent._get_id()
#             re_number_nodes(node)
#         elif isinstance(node, StochasticObject):
#             node.id_ = parent._get_id()
#             re_number_nodes(node)


def add_bond_bonding_descriptor_via_index(
        parent,
        bond_symbol: str,
        descriptor: str,
        index_: int,
        prior_atom: Atom
):
    bd_atom = _get_bonding_descriptor_atom(parent, descriptor, index_)
    bond = add_bond(parent, bond_symbol, prior_atom, bd_atom)

    if prior_atom == prior_atom.parent.nodes[0]:
        parent.nodes.insert(0, bond)
        parent.bonds.insert(0, bond)
        parent.nodes.insert(0, bd_atom)
        parent.atoms.insert(0, bd_atom)
    elif prior_atom == prior_atom.parent.nodes[-1]:
        parent.nodes.append(bond)
        parent.bonds.append(bond)
        parent.nodes.append(bd_atom)
        parent.atoms.append(bd_atom)
    else:
        branch = Branch(parent, parent._get_id())
        branch.nodes = [bond, bd_atom]
        branch.atoms = [bd_atom]
        branch.bond = [bond]
        insert_obj_into_list(parent.nodes, prior_atom, branch)

    insert_obj_into_list(parent.root.bonds, prior_atom.bonds[0], bond)
    insert_obj_into_list(parent.root.atoms, prior_atom, bd_atom)


# def add_stochastic_object(
#         parent: BigSMILES | Branch,
#         left_descriptor: str,
#         left_index: int,
#         right_descriptor: str,
#         right_index: int
# ):
#     stoch_obj = StochasticObject(parent, parent._get_id())
#
#     # create bonding descriptors
#     left_bd = BondDescriptor(stoch_obj, left_descriptor, left_index)
#     right_bd = BondDescriptor(stoch_obj, right_descriptor, right_index)
#     stoch_obj.bonding_descriptors.append(left_bd)
#     stoch_obj.bonding_descriptors.append(right_bd)
#
#     if not stoch_obj.implicit_endgroups:
#         try:
#             prior_atom = get_prior(parent, (Atom, BondDescriptor, StochasticObject))
#         except BigSMILESError:
#             # no prior atoms, it must be the first atom
#             if isinstance(parent, BigSMILES) and:
#                 prior_atom = Atom(parent._get_id(), "H", None, "", 0, 0, 1, parent)
#                 parent.atoms.append(prior_atom)
#
#         bond = add_bond(parent, "", prior_atom, stoch_obj)
#         stoch_obj.bond_left = bond
#
#     parent.nodes.append(stoch_obj)
#
#     return stoch_obj


__all__ = ["ParentType", "add_atom", "add_bond", "add_ring", "add_ring_from_atoms", "add_bonding_descriptor",
           "add_bond_atom_pair", "add_bond_bonding_descriptor_pair", "open_branch", "close_branch",
           "open_stochastic_object", "open_stochastic_object_with_bond", "close_stochastic_object",
           "open_stochastic_fragment", "close_open_stochastic_fragment", "close_stochastic_fragment",
           "exit_", "run_validation"
           ]



"""

Remaining validation done on the bigsmile object

Note:
* tokenization should take care of symbol validation
* Bigsmiles_constructor should take care of most bond and valance issues

This is mainly to catch additional leftover validation.

"""

from bigsmiles.errors import BigSMILESError
from bigsmiles.bigsmiles import BigSMILES, StochasticObject, StochasticFragment, \
    BondDescriptor, Branch


## Pre-construction validation#############################################################################
def pre_validation(text: str):
    """ Main entry point for pre-construction validation. """
    branch_symbol_validation(text)
    stochastic_object_validation(text)
    brackets_validation(text)


def branch_symbol_validation(text: str):
    count = text.count('(') - text.count(')')
    if count != 0:
        if count > 0:
            raise BigSMILESError(f"Invalid BigSMILES. Missing {count} closing branch symbols ')'. "
                                 f"\n Invalid string: {text}")
        raise BigSMILESError(f"Invalid BigSMILES. Missing {count} opening branch symbols '('. "
                             f"\n Invalid string: {text}")


def stochastic_object_validation(text: str):
    count = text.count('{') - text.count('}')
    if count != 0:
        if count > 0:
            raise BigSMILESError(f"Invalid BigSMILES. Missing {count} closing stochastic object symbols" + " '}'. "
                                 f"\n Invalid string: {text}")
        raise BigSMILESError(f"Invalid BigSMILES. Missing {count} opening stochastic object symbols " + "'{'. "
                             f"\n Invalid string: {text}")


def brackets_validation(text: str):
    count = text.count('[') - text.count(']')
    if count != 0:
        if count > 0:
            raise BigSMILESError(f"Invalid BigSMILES. Missing {count} closing bracket symbols ']'. "
                                 f"\n Invalid string: {text}")
        raise BigSMILESError(f"Invalid BigSMILES. Missing {count} opening bracket symbols '['. "
                             f"\n Invalid string: {text}")


## Post-construction validation#############################################################################
def post_validation(bigsmiles: BigSMILES):
    """ Main entry point for post-construction validation. """
    check_ring_closure(bigsmiles)
    check_bonding_descriptors(bigsmiles)
    check_implicit_endgroups_ends(bigsmiles)


def check_ring_closure(bigsmiles: BigSMILES):
    """ Checks to see that all rings have been closed that were started. """
    for ring in bigsmiles.rings:
        if ring.atom2 is None:
            raise BigSMILESError(f"Ring opened, but not closed. (ring id: {ring.ring_id})")


def check_bonding_descriptors(bigsmiles: BigSMILES | StochasticObject | Branch):
    """

    rules:
    if [$] must be 2 or more
    if [>] there must be [<]
    if [<] there must be [>]
    if end is implicit; there must be single bonding units
    if ends are explict they must both have a matching pair in the stochastic fragments



    Parameters
    ----------
    bigsmiles

    Returns
    -------

    """
    for node in bigsmiles.nodes:
        if isinstance(node, StochasticObject):
            do_check_bonding_descriptors(node)

            # check nested stochastic objects
            check_bonding_descriptors(node)
        if isinstance(node, Branch):
            check_bonding_descriptors(node)


def do_check_bonding_descriptors(stoch_obj: StochasticObject):
    for bd in stoch_obj.bonding_descriptors:
        # if [$], [$0], [$1], ... must be 2 or more
        if bd.descriptor == "$":
            if len(bd.instances) <= 1:
                raise BigSMILESError("[$] type bonding descriptors require more than one instances in a string.")

        # if [>] there must be [<] and if [<] there must be [>]
        if bd.descriptor in ("<", ">"):
            bd_pair = find_complementing_bonding_descriptor(stoch_obj, bd)
            if bd_pair is None:
                raise BigSMILESError(f'{bd} complementary partner not found.')


def find_complementing_bonding_descriptor(stoch_obj: StochasticObject, bond_descr: BondDescriptor) \
        -> BondDescriptor | None:
    for bd in stoch_obj.bonding_descriptors:
        if bond_descr.is_pair(bd):
            return bd

    return None


def check_implicit_endgroups_ends(obj: BigSMILES | StochasticObject | StochasticFragment | Branch,
                                  parent_obj: BigSMILES | StochasticObject | StochasticFragment | Branch = None):
    # if end is implicit; there must be single bonding units
    for i, node in enumerate(obj.nodes):
        if isinstance(node, StochasticObject):
            if node.end_group_left.descriptor.descriptor == "":
                if i != 0:
                    # nothing allowed to the left
                    raise BigSMILESError("With a the left end group implicit, "
                                         "there should be nothing to the left of the stochastic object.")
                if parent_obj is not None:
                    # if it isn't BigSMILES it can't have a left implicit end group
                    raise BigSMILESError("Implicit left end group not allowed within interior.")

            if node.end_group_right.descriptor.descriptor == "":

                if i != len(obj.nodes) - 1:
                    # nothing allowed to the right
                    raise BigSMILESError("With a implicit right end group, there should not be anything to the "
                                         "right of it.")

        if hasattr(node, "nodes"):
            check_implicit_endgroups_ends(node, obj)





class BigSMILESError(Exception):
    def __init__(self, text: str, error=None):
        if isinstance(error, BigSMILESError):
            text += '\n\t'
            text += error.text.replace("\t", "\t\t")
        self.text = text

    def __str__(self):
        return self.text




"""

Code for tokenizing a BigSMILES string.

"""

import enum
import re

from bigsmiles.config import Config


class BigSMILESTokenizeError(Exception):
    """ Raised when error tokenizing a BigSMILES string. """
    def __init__(self, text: str):
        self.text = text

    def __str__(self):
        return self.text


class TokenKind(enum.Enum):
    Bond = 0
    Atom = 1
    Aromatic = 2
    AtomExtend = 3
    BranchStart = 4
    BranchEnd = 5
    Ring = 6
    Ring2 = 7
    BondEZ = 8
    Mix = 9
    Rxn = 10
    BondDescriptor = 11
    StochasticSeperator = 12
    StochasticStart = 13
    StochasticEnd = 14
    ImplictEndGroup = 15
    BondDescriptorLadder = 16


_isotope_pattern = r'(?P<isotope>[\d]{1,3})?'
_element_pattern = r'(?P<element>' + "|".join(Config.elements_ordered) + "|".join(Config.aromatic) + '{1})'
_stereo_pattern = r'(?P<stereo>@{1,2})?'
_hydrogen_pattern = r'(?P<hcount>H[\d]?)?'
_charge_pattern = r'(?P<charge>[-|\+]{1,3}[\d]?)?'
atom_pattern = r"(?:\[)" + _isotope_pattern + _element_pattern + _stereo_pattern + \
               _hydrogen_pattern + _charge_pattern + r"(?:\])"

token_specification = [
    # order in the list is important; regex stops at first match
    (TokenKind.Bond.name, r'[=|#]'),
    (TokenKind.Atom.name, "|".join(Config.elements_ordered)),
    (TokenKind.Aromatic.name, "|".join(Config.aromatic)),
    (TokenKind.AtomExtend.name, atom_pattern),  # Atom in brackets
    (TokenKind.BranchStart.name, r'\('),
    (TokenKind.BranchEnd.name, r'\)'),
    (TokenKind.Ring.name, r'[\d]{1}'),
    (TokenKind.Ring2.name, r'%[\d]{2}'),  # ring_id with two-digit numbers
    (TokenKind.BondEZ.name, r'/|\\'),  # cis trans
    (TokenKind.Mix.name, r"\."),  # mixture
    (TokenKind.Rxn.name, r">>|>"),  # reaction -->

    (TokenKind.BondDescriptorLadder.name, r"\[[$<>][\d]\[[$<>][\d]?\][\d]?\]"),  # Ladder
    (TokenKind.BondDescriptor.name, r"\[[$<>][\d]?[\d]?\]"),
    (TokenKind.StochasticSeperator.name, r",|;"),
    (TokenKind.StochasticStart.name, r'\{'),
    (TokenKind.StochasticEnd.name, r'\}'),
    (TokenKind.ImplictEndGroup.name, r'\[\]'),

    ('SKIP', r'[ \t]+'),  # Skip over spaces and tabs
    ('MISMATCH', r'.'),  # Any other character
]

tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_specification)

tok_regex = re.compile(tok_regex)


class Token:
    __slots__ = ("kind", "value")

    def __init__(self, kind: TokenKind, value: str):
        self.kind = kind
        self.value = value

    def __str__(self):
        return f"{self.kind}: {self.value}"


def tokenize(text: str) -> list[Token]:
    """

    Tokenizes a bigSMILES string.

    Parameters
    ----------
    text: str
        BigSMILES string

    Returns
    -------
    result: list[Token]
        BigSMILES as a token list

    """
    result = []
    for match in re.finditer(tok_regex, text.strip()):
        kind = match.lastgroup

        value = match.group()
        if kind == 'SKIP':
            continue
        elif kind == 'MISMATCH':
            raise BigSMILESTokenizeError(f'Invalid symbol (or group of symbols). (starting with {value!r}; '
                                         f'index: {match.span()[0]})'
                                         f'\n{text}' + "\n" + " " * match.span()[0] + "^(and forward)")

        result.append(
            Token(TokenKind[kind], value)
        )

    return result



import warnings


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return str(msg) + '\n'

warnings.formatwarning = custom_formatwarning

from bigsmiles.errors import BigSMILESError
from bigsmiles.validation import pre_validation
from bigsmiles.tokenizer import Token, TokenKind, tokenize
import bigsmiles.constructor_str as constructor
from bigsmiles.bigsmiles import BigSMILES, Branch, BondDescriptorAtom, Bond, Atom, StochasticObject, StochasticFragment


def map_atom(parent: constructor.ParentType, tokens: list[Token], token: Token):
    if isinstance(parent, BigSMILES) and not parent:
        return constructor.add_atom(parent, token.value)

    if isinstance(parent, StochasticFragment):
        return constructor.add_atom(parent, token.value)

    return constructor.add_bond_atom_pair(parent, "", token.value)


def map_bond(parent: constructor.ParentType, tokens: list[Token], token: Token):
    try:
        next_token = tokens.pop(0)
    except IndexError:
        raise BigSMILESError("Bond can't be at the end of a BigSMILES string.")

    if next_token.kind in (TokenKind.Atom, TokenKind.AtomExtend):
        return constructor.add_bond_atom_pair(parent, token.value, next_token.value)

    if next_token.kind is not TokenKind.BondDescriptor:
        return constructor.add_bond_bonding_descriptor_pair_str(parent, token.value, next_token.value)

    return map_stochastic_object_start(parent, tokens, next_token, token)


def map_bond_descriptor(parent: constructor.ParentType, tokens: list[Token], token: Token):
    if tokens[0].kind is TokenKind.StochasticEnd:
        parent = constructor.close_stochastic_fragment_str(parent)
        parent = constructor.close_stochastic_object_str(parent, token.value)
        tokens.pop(0)
        return parent

    if isinstance(parent, StochasticFragment):
        # first StochasticFragment symbol
        return constructor.add_bonding_descriptor_str(parent, token.value)

    if isinstance(parent, BigSMILES) and not parent:
        # first Branch symbol
        # check to make sure the branch closes immediately
        if tokens[0].kind != TokenKind.BranchEnd:
            raise BigSMILESError("If Bonding Descriptors is the first Branch symbol, it can only be followed by "
                                 "Branch End.")

    return constructor.add_bond_bonding_descriptor_pair_str(parent, "", token.value)


def map_branch_start(parent: constructor.ParentType, tokens: list[Token], token: Token):
    if isinstance(parent, BigSMILES) and not parent:
        raise BigSMILESError("BigSMILES can't start with a branch symbol.")
    if isinstance(parent, Branch) and len(parent.nodes) == 0:
        raise BigSMILESError("Branch can't be the first object within a branch.")
    return constructor.open_branch(parent)


def map_branch_end(parent: constructor.ParentType, tokens: list[Token], token: Token):
    return constructor.close_branch(parent)


def map_ring(parent: constructor.ParentType, tokens: list[Token], token: Token):
    return constructor.add_ring(parent, int(token.value.replace('%', '')))


def map_stochastic_object_start(parent: constructor.ParentType, tokens: list[Token], token: Token,
                                bond_token: Token = None):
    try:
        next_token = tokens.pop(0)
    except IndexError:
        raise BigSMILESError("Stochastic objects must begin with a bond descriptor (or implicit bonding descriptor).")

    if next_token.kind not in (TokenKind.ImplictEndGroup, TokenKind.BondDescriptor):
        raise BigSMILESError(f"Stochastic object starts must be followed an explict or implicit end group.")

    if isinstance(parent, BigSMILES) and not parent:
        return constructor.open_stochastic_object_str(parent, next_token.value)
    else:
        if bond_token is None:
            return constructor.open_stochastic_object_with_bond_str(parent, "", next_token.value)
        else:
            return constructor.open_stochastic_object_with_bond_str(parent, bond_token.value, next_token.value)


def map_stochastic_object_end(parent: constructor.ParentType, tokens: list[Token], token: Token) -> TokenKind:
    raise BigSMILESError("Stochastic objects should end with bonding descriptor (or implicit bonding description)")


def map_bond_seperator(parent: constructor.ParentType, tokens: list[Token], token: Token):
    return constructor.close_open_stochastic_fragment_str(parent)


def NotImplementedFunc(*args, **kwargs):
    raise NotImplementedError()


def SkipSymbol(parent: constructor.ParentType, tokens: list[Token], token: Token):
    warnings.warn(f"Symbol skipped: {token.value}")
    return parent


map_tokens = {
    TokenKind.Bond: map_bond,
    TokenKind.Atom: map_atom,
    TokenKind.Aromatic: map_atom,
    TokenKind.AtomExtend: map_atom,
    TokenKind.BranchStart: map_branch_start,
    TokenKind.BranchEnd: map_branch_end,
    TokenKind.Ring: map_ring,
    TokenKind.Ring2: map_ring,
    TokenKind.BondEZ: SkipSymbol,
    TokenKind.Mix: NotImplementedFunc,
    TokenKind.Rxn: NotImplementedFunc,
    TokenKind.BondDescriptor: map_bond_descriptor,
    TokenKind.StochasticSeperator: map_bond_seperator,
    TokenKind.StochasticStart: map_stochastic_object_start,
    TokenKind.StochasticEnd: map_stochastic_object_end,
    TokenKind.ImplictEndGroup: map_bond_descriptor,
    TokenKind.BondDescriptorLadder: NotImplementedFunc
}


def tokens_to_bigsmiles(parent: constructor.ParentType, tokens: list[Token]):
    """

    Main loop for converting tokens into BigSMILES objects.

    Parameters
    ----------
    parent: constructor.ParentType

    tokens:

    Returns
    -------

    """
    num_tokens = len(tokens)
    while tokens:
        token = tokens.pop(0)
        func = map_tokens[token.kind]
        try:
            parent = func(parent, tokens, token)
        except BigSMILESError as e:
            raise BigSMILESError(f"Issue with token '{token}'. (index: {num_tokens-len(tokens)-1})", e)


def parse_bigsmiles_str(input_text: str, bigsmiles):
    """
    Main function that turns BigSMILES string into a BigSMILES object.
    Constructs BigSMILES tree in the provided object.
    """
    pre_validation(input_text)
    tokens = tokenize(input_text)

    try:
        tokens_to_bigsmiles(bigsmiles, tokens)
        constructor.run_validation(bigsmiles)
        constructor.exit_(bigsmiles)
    except BigSMILESError as e:
        raise BigSMILESError(f"Parsing failed on '{input_text}'.", e)




from bigsmiles.bigsmiles import BigSMILES, Branch, StochasticObject, StochasticFragment


def run_syntax_fixes(bigsmiles: BigSMILES):
    if not bigsmiles:
        return

    remove_unnecessary_branch_symbols(bigsmiles)
    check_ring_numbers(bigsmiles)


def remove_unnecessary_branch_symbols(obj: BigSMILES | StochasticObject | StochasticFragment | Branch):
    # check if branch is notation is need; if not needed remove it and re-do ids
    # Example: CCC(C(CC)) --> CCCCCC

    for node in obj.nodes:
        if hasattr(node, "nodes"):
            remove_unnecessary_branch_symbols(node)

    branch_check(obj)


def branch_check(obj: StochasticObject | StochasticFragment | Branch):
    # checks if branch is end of nodes; if it is then it's not needed
    if isinstance(obj.nodes[-1], Branch):
        branch = obj.nodes.pop()
        obj.nodes += branch.nodes

        if hasattr(obj, 'parent'):
            renumber_branches(obj.parent)
        else:
            renumber_branches(obj)


def renumber_branches(obj, counter: int = 0) -> int:
    # Recursive function: loops through obj and re-numbers branch.id_
    for node in obj.nodes:
        if isinstance(node, Branch):
            node.id_ = counter
            counter += 1
        if hasattr(node, "nodes"):
            counter = renumber_branches(node, counter)

    return counter


def check_ring_numbers(obj: BigSMILES):
    # check if ring_counter had been skipped and restart everything at 1
    for i, ring in enumerate(obj.rings):
        ring.ring_id = i + 1



import pkg_resources
__version__ = pkg_resources.require("bigsmiles")[0].version


from bigsmiles.config import Config
from bigsmiles.errors import BigSMILESError
from bigsmiles.tokenizer import BigSMILESTokenizeError
from bigsmiles.bigsmiles import Atom, Bond, Branch, BondDescriptor, StochasticObject, StochasticFragment, BigSMILES

errors = [
    BigSMILESError,
    BigSMILESTokenizeError
]

